@@ -1,10 +1,12 @@
 package com.fasterxml.jackson.dataformat.xml.deser;
 
 import java.io.IOException;
+import java.io.Writer;
 import java.math.BigDecimal;
 import java.math.BigInteger;
 import java.util.Set;
 
+import javax.xml.stream.XMLStreamException;
 import javax.xml.stream.XMLStreamReader;
 import javax.xml.stream.XMLStreamWriter;
 
@@ -14,6 +16,7 @@ import com.fasterxml.jackson.core.io.IOContext;
 import com.fasterxml.jackson.core.util.ByteArrayBuilder;
 import com.fasterxml.jackson.dataformat.xml.PackageVersion;
 import com.fasterxml.jackson.dataformat.xml.XmlMapper;
+import com.fasterxml.jackson.dataformat.xml.util.StaxUtil;
 
 /**
  * {@link JsonParser} implementation that exposes XML structure as
@@ -30,10 +33,21 @@ public class FromXmlParser
 
     /**
      * Enumeration that defines all togglable features for XML parsers.
-     * None defined so far (2.6), so just a placeholder.
      */
     public enum Feature implements FormatFeature
     {
+        /**
+         * Feature that indicates whether XML Empty elements (ones where there are
+         * no separate start and end tages, but just one tag that ends with "/>")
+         * are exposed as {@link JsonToken#VALUE_NULL}) or not. If they are not
+         * returned as `null` tokens, they will be returned as {@link JsonToken#VALUE_STRING}
+         * tokens with textual value of "" (empty String).
+         *<p>
+         * Default setting is `true` for backwards compatibility.
+         *
+         * @since 2.9
+         */
+        EMPTY_ELEMENT_AS_NULL(true)
         ;
 
         final boolean _defaultState;
@@ -170,7 +184,8 @@ public class FromXmlParser
         _parsingContext = XmlReadContext.createRootContext(-1, -1);
         // and thereby start a scope
         _nextToken = JsonToken.START_OBJECT;
-        _xmlTokens = new XmlTokenStream(xmlReader, ctxt.getSourceReference());
+        _xmlTokens = new XmlTokenStream(xmlReader, ctxt.getSourceReference(),
+                _formatFeatures);
     }
 
     @Override
@@ -214,11 +229,13 @@ public class FromXmlParser
 
     public FromXmlParser enable(Feature f) {
         _formatFeatures |= f.getMask();
+        _xmlTokens.setFormatFeatures(_formatFeatures);
         return this;
     }
 
     public FromXmlParser disable(Feature f) {
         _formatFeatures &= ~f.getMask();
+        _xmlTokens.setFormatFeatures(_formatFeatures);
         return this;
     }
 
@@ -358,8 +375,9 @@ public class FromXmlParser
                 } else {
                     _xmlTokens.close();
                 }
+            } catch (XMLStreamException e) {
+                StaxUtil.throwAsParseException(e, this);
             } finally {
-                // as per [JACKSON-324], do in finally block
                 // Also, internal buffer(s) can now be released as well
                 _releaseBuffers();
             }
@@ -477,8 +495,12 @@ public class FromXmlParser
             }
             return t;
         }
-        int token = _xmlTokens.next();
-
+        int token;
+        try {
+            token = _xmlTokens.next();
+        } catch (XMLStreamException e) {
+            token = StaxUtil.throwAsParseException(e, this);
+        }
         // Need to have a loop just because we may have to eat/convert
         // a start-element that indicates an array element.
         while (token == XmlTokenStream.XML_START_ELEMENT) {
@@ -492,7 +514,11 @@ public class FromXmlParser
             if (_parsingContext.inArray()) {
                 // Yup: in array, so this element could be verified; but it won't be
                 // reported anyway, and we need to process following event.
-                token = _xmlTokens.next();
+                try {
+                    token = _xmlTokens.next();
+                } catch (XMLStreamException e) {
+                    StaxUtil.throwAsParseException(e, this);
+                }
                 _mayBeLeaf = true;
                 continue;
             }
@@ -554,7 +580,11 @@ public class FromXmlParser
                     // we had an empty String (or all white space), and we are
                     // deserializing an array, we better hide the empty text.
                     // Also: must skip following END_ELEMENT
-                    _xmlTokens.skipEndElement();
+                    try {
+                        _xmlTokens.skipEndElement();
+                    } catch (XMLStreamException e) {
+                        StaxUtil.throwAsParseException(e, this);
+                    }
                     if (_parsingContext.inArray()) {
                         if (_isEmpty(_currText)) {
                             // 06-Jan-2015, tatu: as per [dataformat-xml#180], need to
@@ -572,7 +602,11 @@ public class FromXmlParser
                     // loop over again
                     if (_parsingContext.inObject()) {
                         if ((_currToken != JsonToken.FIELD_NAME) && _isEmpty(_currText)) {
-                            token = _xmlTokens.next();
+                            try {
+                                token = _xmlTokens.next();
+                            } catch (XMLStreamException e) {
+                                StaxUtil.throwAsParseException(e, this);
+                            }
                             continue;
                         }
                     }
@@ -586,7 +620,7 @@ public class FromXmlParser
             }
         }
     }
-    
+
     /*
     /**********************************************************
     /* Overrides of specialized nextXxx() methods
@@ -614,7 +648,13 @@ public class FromXmlParser
             return null;
         }
 
-        int token = _xmlTokens.next();
+        int token;
+
+        try {
+            token = _xmlTokens.next();
+        } catch (XMLStreamException e) {
+            token = StaxUtil.throwAsParseException(e, this);
+        }
 
         // mostly copied from 'nextToken()'
         while (token == XmlTokenStream.XML_START_ELEMENT) {
@@ -625,7 +665,11 @@ public class FromXmlParser
                 return null;
             }
             if (_parsingContext.inArray()) {
-                token = _xmlTokens.next();
+                try {
+                    token = _xmlTokens.next();
+                } catch (XMLStreamException e) {
+                    StaxUtil.throwAsParseException(e, this);
+                }
                 _mayBeLeaf = true;
                 continue;
             }
@@ -652,7 +696,6 @@ public class FromXmlParser
             _parsingContext = _parsingContext.getParent();
             _namesToWrap = _parsingContext.getNamesToWrap();
             break;
-            
         case XmlTokenStream.XML_ATTRIBUTE_NAME:
             // If there was a chance of leaf node, no more...
             if (_mayBeLeaf) {
@@ -667,16 +710,18 @@ public class FromXmlParser
             }
             break;
         case XmlTokenStream.XML_ATTRIBUTE_VALUE:
-            _currText = _xmlTokens.getText();
             _currToken = JsonToken.VALUE_STRING;
-            break;
+            return (_currText = _xmlTokens.getText());
         case XmlTokenStream.XML_TEXT:
             _currText = _xmlTokens.getText();
             if (_mayBeLeaf) {
                 _mayBeLeaf = false;
                 // Also: must skip following END_ELEMENT
-                _xmlTokens.skipEndElement();
-
+                try {
+                    _xmlTokens.skipEndElement();
+                } catch (XMLStreamException e) {
+                    StaxUtil.throwAsParseException(e, this);
+                }
                 // NOTE: this is different from nextToken() -- NO work-around
                 // for otherwise empty List/array
                 _currToken = JsonToken.VALUE_STRING;
@@ -759,7 +804,7 @@ public class FromXmlParser
             // the interesting case; may be able to convert certain kinds of
             // elements (specifically, ones with attributes, CDATA only content)
             // into VALUE_STRING
-            {
+            try {
                 String str = _xmlTokens.convertToString();
                 if (str != null) {
                     // need to convert token, as well as "undo" START_OBJECT
@@ -769,13 +814,16 @@ public class FromXmlParser
                     _namesToWrap = _parsingContext.getNamesToWrap();
                     _currToken = JsonToken.VALUE_STRING;
                     _nextToken = null;
-                    /* One more thing: must explicitly skip the END_OBJECT that
-                     * would follow.
-                     */
-                    _xmlTokens.skipEndElement();
+                    // One more thing: must explicitly skip the END_OBJECT that would follow
+                    try {
+                        _xmlTokens.skipEndElement();
+                    } catch (XMLStreamException e) {
+                        StaxUtil.throwAsParseException(e, this);
+                    }
                     return (_currText = str);
                 }
-                
+            } catch (XMLStreamException e) {
+                StaxUtil.throwAsParseException(e, this);
             }
             return null;
         default:
@@ -813,6 +861,17 @@ public class FromXmlParser
         return false;
     }
 
+    @Override // since 2.8
+    public int getText(Writer writer) throws IOException
+    {
+        String str = getText();
+        if (str == null) {
+            return 0;
+        }
+        writer.write(str);
+        return str.length();
+    }
+    
     /*
     /**********************************************************
     /* Public API, access to token information, binary
@@ -923,10 +982,15 @@ public class FromXmlParser
     protected void _handleEOF() throws JsonParseException
     {
         if (!_parsingContext.inRoot()) {
-            _reportInvalidEOF(": expected close marker for "+_parsingContext.getTypeDesc()+" (from "+_parsingContext.getStartLocation(_ioContext.getSourceReference())+")");
+            String marker = _parsingContext.inArray() ? "Array" : "Object";
+            _reportInvalidEOF(String.format(
+                    ": expected close marker for %s (start marker at %s)",
+                    marker,
+                    _parsingContext.getStartLocation(_ioContext.getSourceReference())),
+                    null);
         }
     }
-    
+
     /*
     /**********************************************************
     /* Internal methods
