@@ -8,7 +8,6 @@ import org.codehaus.stax2.XMLStreamReader2;
 import org.codehaus.stax2.ri.Stax2ReaderAdapter;
 
 import com.fasterxml.jackson.core.JsonLocation;
-import com.fasterxml.jackson.dataformat.xml.util.StaxUtil;
 
 /**
  * Simple helper class used on top of STAX {@link XMLStreamReader} to further
@@ -43,6 +42,13 @@ public class XmlTokenStream
     final protected XMLStreamReader2 _xmlReader;
 
     final protected Object _sourceReference;
+
+    /**
+     * Bit flag composed of bits that indicate which
+     * {@link FromXmlParser.Feature}s
+     * are enabled.
+     */
+    protected int _formatFeatures;
     
     /*
     /**********************************************************************
@@ -53,7 +59,14 @@ public class XmlTokenStream
     protected int _currentState;
 
     protected int _attributeCount;
-    
+
+    /**
+     * If true we have a START_ELEMENT with mixed text
+     *
+     * @since 2.8
+     */
+    protected boolean _mixedText;
+
     /**
      * Index of the next attribute of the current START_ELEMENT
      * to return (as field name and value pair), if any; -1
@@ -97,7 +110,8 @@ public class XmlTokenStream
     /**********************************************************************
      */
 
-    public XmlTokenStream(XMLStreamReader xmlReader, Object sourceRef)
+    public XmlTokenStream(XMLStreamReader xmlReader, Object sourceRef,
+            int formatFeatures)
     {
         _sourceReference = sourceRef;
         // Let's ensure we point to START_ELEMENT...
@@ -110,12 +124,20 @@ public class XmlTokenStream
         _localName = _xmlReader.getLocalName();
         _namespaceURI = _xmlReader.getNamespaceURI();
         _attributeCount = _xmlReader.getAttributeCount();
+        _formatFeatures = formatFeatures;
     }
 
     public XMLStreamReader2 getXmlReader() {
         return _xmlReader;
     }
 
+    /**
+     * @since 2.9
+     */
+    protected void setFormatFeatures(int f) {
+        _formatFeatures = f;
+    }
+    
     /*
     /**********************************************************************
     /* Public API
@@ -153,20 +175,15 @@ public class XmlTokenStream
     }
     */
 
-    public int next() throws IOException 
+    public int next() throws XMLStreamException 
     {
         if (_repeatElement != 0) {
             return (_currentState = _handleRepeatElement());
         }
-        try {
-            return _next();
-        } catch (XMLStreamException e) {
-            StaxUtil.throwXmlAsIOException(e);
-            return -1;
-        }
+        return _next();
     }
-    
-    public void skipEndElement() throws IOException
+
+    public void skipEndElement() throws IOException, XMLStreamException
     {
         int type = next();
         if (type != XML_END_ELEMENT) {
@@ -183,27 +200,18 @@ public class XmlTokenStream
         return (_currentState == XML_START_ELEMENT) && (_attributeCount > 0);
     }
     
-    public void closeCompletely() throws IOException
-    {
-        try {
-            _xmlReader.closeCompletely();
-        } catch (XMLStreamException e) {
-            StaxUtil.throwXmlAsIOException(e);
-        }
+    public void closeCompletely() throws XMLStreamException {
+        _xmlReader.closeCompletely();
     }
 
-    public void close() throws IOException
-    {
-        try {
-            _xmlReader.close();
-        } catch (XMLStreamException e) {
-            StaxUtil.throwXmlAsIOException(e);
-        }
+    public void close() throws XMLStreamException {
+        _xmlReader.close();
     }
 
     public JsonLocation getCurrentLocation() {
         return _extractLocation(_xmlReader.getLocationInfo().getCurrentLocation());
     }
+
     public JsonLocation getTokenLocation() {
         return _extractLocation(_xmlReader.getLocationInfo().getStartLocation());
     }
@@ -262,38 +270,34 @@ public class XmlTokenStream
         }
     }
 
-    protected String convertToString() throws IOException
+    protected String convertToString() throws XMLStreamException
     {
         // only applicable to cases where START_OBJECT was induced by attributes
         if (_currentState != XML_ATTRIBUTE_NAME || _nextAttributeIndex != 0) {
             return null;
         }
-        try {
-            String text = _collectUntilTag();
-            // 23-Dec-2015, tatu: Used to require text not to be null, but as per
-            //   [dataformat-xml#167], empty tag does count
-            if (_xmlReader.getEventType() == XMLStreamReader.END_ELEMENT) {
-                if (text == null) {
-                    text = "";
-                }
-                if (_currentWrapper != null) {
-                    _currentWrapper = _currentWrapper.getParent();
-                }
-                // just for diagnostics, reset to element name (from first attribute name)
-                _localName = _xmlReader.getLocalName();
-                _namespaceURI = _xmlReader.getNamespaceURI();
-                _attributeCount = 0;
-                _currentState = XML_TEXT;
-                _textValue = text;
-                return text;
+        String text = _collectUntilTag();
+        // 23-Dec-2015, tatu: Used to require text not to be null, but as per
+        //   [dataformat-xml#167], empty tag does count
+        if (_xmlReader.getEventType() == XMLStreamReader.END_ELEMENT) {
+            if (text == null) {
+                text = "";
             }
-        } catch (XMLStreamException e) {
-            StaxUtil.throwXmlAsIOException(e);
+            if (_currentWrapper != null) {
+                _currentWrapper = _currentWrapper.getParent();
+            }
+            // just for diagnostics, reset to element name (from first attribute name)
+            _localName = _xmlReader.getLocalName();
+            _namespaceURI = _xmlReader.getNamespaceURI();
+            _attributeCount = 0;
+            _currentState = XML_TEXT;
+            _textValue = text;
+            return text;
         }
         // Anything to do in failed case? Roll back whatever we found or.. ?
         return null;
     }
-    
+
     /*
     /**********************************************************************
     /* Internal methods, parsing
@@ -315,21 +319,36 @@ public class XmlTokenStream
             }
             // otherwise need to find START/END_ELEMENT or text
             String text = _collectUntilTag();
-            // If it's START_ELEMENT, ignore any text
-            if (_xmlReader.getEventType() == XMLStreamReader.START_ELEMENT) {
-                return _initStartElement();
-            }
-            // For END_ELEMENT we will return text, if any
-            if (text != null) {
+            final boolean startElementNext = _xmlReader.getEventType() == XMLStreamReader.START_ELEMENT;
+            // If we have no/all-whitespace text followed by START_ELEMENT, ignore text
+            if (startElementNext) {
+                if (text == null || _allWs(text)) {
+                    _mixedText = false;
+                    return _initStartElement();
+                }
+                _mixedText = true;
                 _textValue = text;
                 return (_currentState = XML_TEXT);
             }
+            // For END_ELEMENT we will return text, if any
+            if (text != null) {
+                _mixedText = false;
+                _textValue = text;
+                return (_currentState = XML_TEXT);
+            }
+            _mixedText = false;
             return _handleEndElement();
+
         case XML_ATTRIBUTE_NAME:
             // if we just returned name, will need to just send value next
             return (_currentState = XML_ATTRIBUTE_VALUE);
         case XML_TEXT:
-            // text is always followed by END_ELEMENT
+            // mixed text with other elements
+            if (_mixedText){
+                _mixedText = false;
+                return _initStartElement();
+            }
+            // text followed by END_ELEMENT
             return _handleEndElement();
         case XML_END:
             return XML_END;
@@ -346,23 +365,54 @@ public class XmlTokenStream
         // START_ELEMENT...
         return _initStartElement();
     }
-    
+
     private final String _collectUntilTag() throws XMLStreamException
     {
-        String text = null;
+        // 21-Jun-2017, tatu: Whether exposed as `null` or "" is now configurable...
+        if (_xmlReader.isEmptyElement()) {
+            _xmlReader.next();
+            if (FromXmlParser.Feature.EMPTY_ELEMENT_AS_NULL.enabledIn(_formatFeatures)) {
+                return null;
+            }
+            return "";
+        }
+
+        CharSequence chars = null;
         while (true) {
             switch (_xmlReader.next()) {
             case XMLStreamConstants.START_ELEMENT:
+                return (chars == null) ? "" : chars.toString();
+
             case XMLStreamConstants.END_ELEMENT:
             case XMLStreamConstants.END_DOCUMENT:
-                return text;
-                // note: SPACE is ignorable (and seldom seen), not to be included
+                // 04-May-2018, tatu: We could easily make <tag></tag> ALSO report
+                //    as `null`, by below, but that breaks existing tests so not
+                //    done at least until 3.0.
+                /*
+                if (chars == null) {
+                    if (FromXmlParser.Feature.EMPTY_ELEMENT_AS_NULL.enabledIn(_formatFeatures)) {
+                        return null;
+                    }
+                    return "";
+                }
+                return chars;
+                */
+                return (chars == null) ? "" : chars.toString();
+
+            // note: SPACE is ignorable (and seldom seen), not to be included
             case XMLStreamConstants.CHARACTERS:
             case XMLStreamConstants.CDATA:
-                if (text == null) {
-                    text = _xmlReader.getText();
-                } else { // can be optimized in future, if need be:
-                    text += _xmlReader.getText();
+                // 17-Jul-2017, tatu: as per [dataformat-xml#236], need to try to...
+                {
+                    String str = _getText(_xmlReader);
+                    if (chars == null) {
+                        chars = str;
+                    } else  {
+                        if (chars instanceof String) {
+                            chars = new StringBuilder(chars);
+                        }
+                        ((StringBuilder)chars).append(str);
+                    }
                 }
                 break;
             default:
@@ -386,7 +436,20 @@ public class XmlTokenStream
         }
         throw new IllegalStateException("Expected to find a tag, instead reached end of input");
     }
-    
+
+    private final String _getText(XMLStreamReader2 r) throws XMLStreamException
+    {
+        try {
+            return r.getText();
+        } catch (RuntimeException e) {
+            Throwable cause = e.getCause();
+            if (cause instanceof XMLStreamException) {
+                throw (XMLStreamException) cause;
+            }
+            throw e;
+        }
+    }
+
     /*
     /**********************************************************************
     /* Internal methods, other
@@ -430,7 +493,7 @@ public class XmlTokenStream
      * Method called to handle details of repeating "virtual"
      * start/end elements, needed for handling 'unwrapped' lists.
      */
-    protected int _handleRepeatElement() throws IOException 
+    protected int _handleRepeatElement() throws XMLStreamException 
     {
         int type = _repeatElement;
         _repeatElement = 0;
@@ -494,22 +557,27 @@ public class XmlTokenStream
                 location.getColumnNumber());
     }
 
+
+    protected boolean _allWs(String str)
+    {
+        final int len = (str == null) ? 0 : str.length();
+        if (len > 0) {
+            for (int i = 0; i < len; ++i) {
+                if (str.charAt(i) > ' ') {
+                    return false;
+                }
+            }
+        }
+        return true;
+    }
+
     // for DEBUGGING
     @Override
     public String toString()
     {
-        StringBuilder sb = new StringBuilder();
-        sb.append("(Token stream:");
-        sb.append(" state=").append(_currentState);
-        sb.append(" attr#=").append(_attributeCount);
-        sb.append(" nextAttr#=").append(_nextAttributeIndex);
-        sb.append(" name=").append(_localName);
-        sb.append(" text=").append(_textValue);
-        sb.append(" repeat?=").append(_repeatElement);
-        sb.append(" wrapper=[").append(_currentWrapper);
-        sb.append("] repeatElement=").append(_repeatElement);
-        sb.append(" nextName=").append(_nextLocalName);
-        sb.append(")");
-        return sb.toString();
+        return String.format("(Token stream: state=%s attr=%s nextAttr=%s"
+                +" name=%s text=%s repeat?=%s wrapper=[%s] repeatElement=%s nextName=%s)",
+                _currentState, _attributeCount, _nextAttributeIndex,
+                _localName, _textValue, _repeatElement, _currentWrapper, _repeatElement, _nextLocalName);
     }
 }
